@@ -1,6 +1,6 @@
 #
 # This file is part of VIRL 2
-# Copyright (c) 2019-2023, Cisco Systems, Inc.
+# Copyright (c) 2019-2024, Cisco Systems, Inc.
 # All rights reserved.
 #
 # Python bindings for the Cisco VIRL 2 Network Simulation Platform
@@ -22,7 +22,10 @@ from __future__ import annotations
 
 import logging
 import time
-from typing import TYPE_CHECKING, Any, Optional
+import warnings
+from typing import TYPE_CHECKING, Any
+
+from ..utils import get_url_from_template
 
 if TYPE_CHECKING:
     import httpx
@@ -34,34 +37,55 @@ DEFAULT_PROXY_PORT = None
 
 
 class Licensing:
+    _URL_TEMPLATES = {
+        "licensing": "licensing",
+        "tech_support": "licensing/tech_support",
+        "authorization_renew": "licensing/authorization/renew",
+        "transport": "licensing/transport",
+        "product_license": "licensing/product_license",
+        "certificate": "licensing/certificate",
+        "registration": "licensing/registration",
+        "registration_renew": "licensing/registration/renew",
+        "deregistration": "licensing/deregistration",
+        "features": "licensing/features",
+        "reservation_action": "licensing/reservation/{action}",
+    }
     max_wait = 30
     wait_interval = 1.5
 
-    def __init__(self, session: httpx.Client) -> None:
-        self._session = session
+    def __init__(self, session: httpx.Client, is_cert_deprecated: bool) -> None:
+        """
+        Manage licensing.
 
-    @property
-    def base_url(self) -> str:
-        return "licensing"
+        :param session: The httpx-based HTTP client for this session with the server.
+        :param is_cert_deprecated: Whether the certificate supported is deprecated.
+        """
+        self._session = session
+        self._is_cert_deprecated = is_cert_deprecated
+
+    def _url_for(self, endpoint, **kwargs):
+        """
+        Generate the URL for a given API endpoint.
+
+        :param endpoint: The desired endpoint.
+        :param **kwargs: Keyword arguments used to format the URL.
+        :returns: The formatted URL.
+        """
+        return get_url_from_template(endpoint, self._URL_TEMPLATES, kwargs)
 
     def status(self) -> dict[str, Any]:
-        """
-        Get current licensing configuration and status.
-        """
-        return self._session.get(self.base_url).json()
+        """Return current licensing configuration and status."""
+        url = self._url_for("licensing")
+        return self._session.get(url).json()
 
     def tech_support(self) -> str:
-        """
-        Get current licensing tech support.
-        """
-        url = self.base_url + "/tech_support"
+        """Return current licensing tech support."""
+        url = self._url_for("tech_support")
         return self._session.get(url).text
 
     def renew_authorization(self) -> bool:
-        """
-        Renew licensing authorization with the backend.
-        """
-        url = self.base_url + "/authorization/renew"
+        """Renew licensing authorization with the backend."""
+        url = self._url_for("authorization_renew")
         response = self._session.put(url)
         _LOGGER.info("The agent has scheduled an authorization renewal.")
         return response.status_code == 204
@@ -69,22 +93,18 @@ class Licensing:
     def set_transport(
         self,
         ssms: str,
-        proxy_server: Optional[str] = None,
-        proxy_port: Optional[int] = None,
+        proxy_server: str | None = None,
+        proxy_port: int | None = None,
     ) -> bool:
-        """
-        Setup licensing transport configuration.
-        """
-        url = self.base_url + "/transport"
+        """Setup licensing transport configuration."""
+        url = self._url_for("transport")
         data = {"ssms": ssms, "proxy": {"server": proxy_server, "port": proxy_port}}
         response = self._session.put(url, json=data)
-        _LOGGER.info("The transport configuration has been accepted. Config: %s.", data)
+        _LOGGER.info(f"The transport configuration has been accepted. Config: {data}.")
         return response.status_code == 204
 
     def set_default_transport(self) -> bool:
-        """
-        Setup licensing transport configuration to default values.
-        """
+        """Setup licensing transport configuration to default values."""
         default_ssms = self.status()["transport"]["default_ssms"]
         return self.set_transport(
             ssms=default_ssms,
@@ -93,19 +113,25 @@ class Licensing:
         )
 
     def set_product_license(self, product_license: str) -> bool:
-        """
-        Setup a product license.
-        """
-        url = self.base_url + "/product_license"
+        """Setup a product license."""
+        url = self._url_for("product_license")
         response = self._session.put(url, json=product_license)
         _LOGGER.info("Product license was accepted by the agent.")
         return response.status_code == 204
 
-    def get_certificate(self) -> Optional[str]:
+    def get_certificate(self) -> str | None:
         """
+        DEPRECATED: There is no replacement as the certificate support was dropped.
+        (Reason: the certificate support was dropped in CML 2.7.0)
+
         Get the currently installed licensing public certificate.
         """
-        url = self.base_url + "/certificate"
+        if self._is_cert_deprecated:
+            warnings.warn(
+                "'Licensing.get_certificate' is deprecated.", DeprecationWarning
+            )
+            return None
+        url = self._url_for("certificate")
         response = self._session.get(url)
         if response.is_success:
             _LOGGER.info("Certificate received.")
@@ -114,29 +140,43 @@ class Licensing:
 
     def install_certificate(self, cert: str) -> bool:
         """
+        DEPRECATED: There is no replacement as the certificate support was dropped.
+        (Reason: the certificate support was dropped in CML 2.7.0)
+
         Set up a licensing public certificate for internal deployment
         of an unregistered product instance.
         """
-        url = self.base_url + "/certificate"
+        if self._is_cert_deprecated:
+            warnings.warn(
+                "'Licensing.install_certificate' is deprecated.", DeprecationWarning
+            )
+            return False
+        url = self._url_for("certificate")
         response = self._session.post(url, content=cert)
         _LOGGER.info("Certificate was accepted by the agent.")
         return response.status_code == 204
 
     def remove_certificate(self) -> bool:
         """
+        DEPRECATED: There is no replacement as the certificate support was dropped.
+        (Reason: the certificate support was dropped in CML 2.7.0)
+
         Clear any licensing public certificate for internal deployment
         of an unregistered product instance.
         """
-        url = self.base_url + "/certificate"
+        if self._is_cert_deprecated:
+            warnings.warn(
+                "'Licensing.remove_certificate' is deprecated.", DeprecationWarning
+            )
+            return False
+        url = self._url_for("certificate")
         response = self._session.delete(url)
         _LOGGER.info("Certificate was removed.")
         return response.status_code == 204
 
     def register(self, token: str, reregister=False) -> bool:
-        """
-        Setup licensing registration.
-        """
-        url = self.base_url + "/registration"
+        """Setup licensing registration."""
+        url = self._url_for("registration")
         response = self._session.post(
             url, json={"token": token, "reregister": reregister}
         )
@@ -144,10 +184,8 @@ class Licensing:
         return response.status_code == 204
 
     def register_renew(self) -> bool:
-        """
-        Request a renewal of licensing registration against current SSMS.
-        """
-        url = self.base_url + "/registration/renew"
+        """Request a renewal of licensing registration against current SSMS."""
+        url = self._url_for("registration_renew")
         response = self._session.put(url)
         _LOGGER.info("The renewal request was accepted by the agent.")
         return response.status_code == 204
@@ -163,10 +201,8 @@ class Licensing:
         return res
 
     def deregister(self) -> int:
-        """
-        Request deregistration from the current SSMS.
-        """
-        url = self.base_url + "/deregistration"
+        """Request deregistration from the current SSMS."""
+        url = self._url_for("deregistration")
         response = self._session.delete(url)
         if response.status_code == 202:
             _LOGGER.warning(
@@ -174,7 +210,6 @@ class Licensing:
                 "unable to deregister from Smart Software Licensing due to a "
                 "communication timeout."
             )
-            # TODO try to register again and unregister
         if response.status_code == 204:
             _LOGGER.info(
                 "The Product Instance was successfully deregistered from Smart "
@@ -183,74 +218,54 @@ class Licensing:
         return response.status_code
 
     def features(self) -> list[dict[str, str | int]]:
-        """
-        Get current licensing features.
-        """
-        url = self.base_url + "/features"
+        """Get current licensing features."""
+        url = self._url_for("features")
         return self._session.get(url).json()
 
     def update_features(self, features: dict[str, int] | list[dict[str, int]]) -> None:
-        """
-        Update licensing feature's explicit count in reservation mode.
-        """
-        url = self.base_url + "/features"
+        """Update licensing feature's explicit count in reservation mode."""
+        url = self._url_for("features")
         self._session.patch(url, json=features)
 
     def reservation_mode(self, data: bool) -> None:
-        """
-        Enable or disable reservation mode in unregistered agent.
-        """
-        url = self.base_url + "/reservation/mode"
+        """Enable or disable reservation mode in unregistered agent."""
+        url = self._url_for("reservation_action", action="mode")
         self._session.put(url, json=data)
         msg = "enabled" if data else "disabled"
-        _LOGGER.info("The reservation mode has been %s.", msg)
+        _LOGGER.info(f"The reservation mode has been {msg}.")
 
     def enable_reservation_mode(self) -> None:
-        """
-        Enable reservation mode in unregistered agent.
-        """
+        """Enable reservation mode in unregistered agent."""
         return self.reservation_mode(data=True)
 
     def disable_reservation_mode(self) -> None:
-        """
-        Disable reservation mode in unregistered agent.
-        """
+        """Disable reservation mode in unregistered agent."""
         return self.reservation_mode(data=False)
 
     def request_reservation(self) -> str:
-        """
-        Initiate reservation by generating request code and message to the user.
-        """
-        url = self.base_url + "/reservation/request"
+        """Initiate reservation by generating request code and message to the user."""
+        url = self._url_for("reservation_action", action="request")
         response = self._session.post(url)
         _LOGGER.info("Reservation request code received.")
         return response.json()
 
     def complete_reservation(self, authorization_code: str) -> str:
-        """
-        Complete reservation by installing authorization code from SSMS.
-        """
-        # TODO
-        url = self.base_url + "/reservation/complete"
+        """Complete reservation by installing authorization code from SSMS."""
+        url = self._url_for("reservation_action", action="complete")
         response = self._session.post(url, json=authorization_code)
         _LOGGER.info("The confirmation code of completed reservation received.")
         return response.json()
 
     def cancel_reservation(self) -> bool:
-        """
-        Cancel reservation request without completing it.
-        """
-        url = self.base_url + "/reservation/cancel"
+        """Cancel reservation request without completing it."""
+        url = self._url_for("reservation_action", action="cancel")
         response = self._session.delete(url)
         _LOGGER.info("The reservation request has been cancelled.")
         return response.status_code == 204
 
     def release_reservation(self) -> str:
-        """
-        Return a completed reservation.
-        """
-        # TODO
-        url = self.base_url + "/reservation/release"
+        """Return a completed reservation."""
+        url = self._url_for("reservation_action", action="release")
         response = self._session.delete(url)
         _LOGGER.info("The return code of the released reservation received.")
         return response.json()
@@ -260,8 +275,7 @@ class Licensing:
         Discard a reservation authorization code for an already cancelled
         reservation request.
         """
-        # TODO
-        url = self.base_url + "/reservation/discard"
+        url = self._url_for("reservation_action", action="discard")
         response = self._session.post(url, json=data)
         _LOGGER.info(
             "The discard code for an already cancelled reservation request received."
@@ -269,37 +283,29 @@ class Licensing:
         return response.json()
 
     def get_reservation_confirmation_code(self) -> str:
-        """
-        Get the confirmation code.
-        """
-        url = self.base_url + "/reservation/confirmation_code"
+        """Get the reservation confirmation code."""
+        url = self._url_for("reservation_action", action="confirmation_code")
         response = self._session.get(url)
         _LOGGER.info("The confirmation code of the completed reservation received.")
         return response.json()
 
     def delete_reservation_confirmation_code(self) -> bool:
-        """
-        Remove the confirmation code.
-        """
-        url = self.base_url + "/reservation/confirmation_code"
+        """Remove the reservation confirmation code."""
+        url = self._url_for("reservation_action", action="confirmation_code")
         response = self._session.delete(url)
         _LOGGER.info("The confirmation code has been removed.")
         return response.status_code == 204
 
     def get_reservation_return_code(self) -> str:
-        """
-        Get the return code.
-        """
-        url = self.base_url + "/reservation/return_code"
+        """Get the reservation return code."""
+        url = self._url_for("reservation_action", action="return_code")
         response = self._session.get(url)
         _LOGGER.info("The return code of the released reservation received.")
         return response.json()
 
     def delete_reservation_return_code(self) -> bool:
-        """
-        Remove the return code.
-        """
-        url = self.base_url + "/reservation/return_code"
+        """Remove the reservation return code."""
+        url = self._url_for("reservation_action", action="return_code")
         response = self._session.delete(url)
         _LOGGER.info("The return code has been removed.")
         return response.status_code == 204
@@ -319,9 +325,9 @@ class Licensing:
             if count > self.max_wait:
                 timeout = self.max_wait * self.wait_interval
                 raise RuntimeError(
-                    "Timeout: licensing {} did not reach {} status after {} secs. "
-                    "Last status was {}".format(what, target_status, timeout, status)
+                    f"Timeout: licensing {what} did not reach {target_status} status "
+                    f"after {timeout} secs. Last status was {status}"
                 )
             status = self.status()[what]["status"]
-            _LOGGER.debug("%s status: %s", what, status)
+            _LOGGER.debug(f"{what} status: {status}")
             count += 1
